from sidecar.const import Const
from sidecar.utils import Utils
from sidecar.status_maintainer import StatusMaintainer
from sidecar.kub_api_service import IKubApiService
from sidecar.sandbox_error import SandboxError
from logging import Logger
from typing import List
import json


class KUBStatusMaintainer(StatusMaintainer):
    def __init__(self, logger: Logger, kub_api_service: IKubApiService):
        super().__init__(logger)
        self.kub_api_service = kub_api_service

    # annotation is like an item in a dynamoDB
    def update_annotation_in_namespace(self, key: str, value: str):
        annotation = {key: value}
        self.kub_api_service.update_namespace(annotation)

    def update_lazy_load_artifacts_waiting(self, value: bool):
        self.update_annotation_in_namespace(Const.LAZY_LOAD_ARTIFACTS_WAITING, str(value))

    def update_sandbox_errors(self, errors: List[SandboxError]):
        errors_to_db = [{"message": error.message, "code": error.code, "time": error.time} for error in errors]
        self.update_annotation_in_namespace(Const.SANDBOX_ERRORS, json.dumps(errors_to_db))

    def update_qualiy_status(self, status: str):
        self.update_annotation_in_namespace(Const.QUALIY_STATUS, status)

    def get_sandbox_errors(self) -> List[SandboxError]:
        try:
            errors_in_jason_format = self.kub_api_service.get_json_annotation(Const.SANDBOX_ERRORS)
            if errors_in_jason_format is None:
                return []

            errors = []
            for error in errors_in_jason_format:
                errors.append(SandboxError(message=str(error["message"]),
                                           code=str(error["code"]),
                                           time=str(error["time"])))
            return errors

        except Exception as e:
            return []




import os
from os.path import expanduser

from sidecar.const import Const


class KubTokenProvider:
    def get_token(self):
        with open(Const.get_kubernetes_token_file_path(), 'r') as pod_token_file:
            return pod_token_file.read().replace('\n', '')


class FakeKubTokenProvider(KubTokenProvider):
    def get_token(self):
        with open(os.path.join(expanduser("~"), 'sidecar/kub_token',), 'r') as token_file:
            return token_file.read().replace('\n', '')




class AzureInstancePowerState:
    RUNNING = 'PowerState/running'
    DEALLOCATED = 'PowerState/deallocated'
    STOPPED = 'PowerState/stopped'
    STARTING = 'PowerState/starting'
    DEALLOCATING = 'PowerState/deallocating'
    STOPPING = 'PowerState/stopping'



import datetime
import inspect
import logging
import sys
from logging import Logger

import requests
import time

from sidecar.const import Const


class Utils:
    @staticmethod
    def str_to_bool(value: str) -> bool:
        if value.lower() == 'true':
            return True
        else:
            return False

    @staticmethod
    def stop_on_debug():
        while not sys.gettrace():
            time.sleep(0.5)

    @staticmethod
    def read_log(app_name: str) -> str:
        file_path = Const.get_app_log_file(app_name)
        with open(file_path, 'r') as application_log:
            return application_log.read()

    @staticmethod
    def get_timestamp():
        return datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    @staticmethod
    def wait_for(func, interval_sec: int = 10, max_retries: int = 10, error: str = "", silent: bool = False):
        current_retry = 0
        while current_retry < max_retries:
            try:
                if func():
                    return
                else:
                    if not silent:
                        print('re-try {} out of {}'.format(current_retry, max_retries))
                    time.sleep(interval_sec)
                    current_retry += 1
                    if current_retry >= max_retries:
                        raise Exception('max retries for wait_for is exhausted with message: {}'.format(error))
            except Exception as e:
                raise Exception('wait_for function exited due to an exception: {}'.format(e))

    @staticmethod
    def retry_on_exception(func,
                           logger: Logger = None,
                           logger_msg: str = "",
                           interval: int = 1,
                           timeout_in_sec: int = 10):
        start = datetime.datetime.now()
        attempt_number = 1
        while True:
            try:
                res = func()
                if attempt_number > 1:
                    if logger:
                        elapsed = datetime.datetime.now() - start
                        logger.info(
                            "{} (succeed after {}s with {} attempts)".format(logger_msg, elapsed.total_seconds(),
                                                                             attempt_number))
                return res
            except requests.exceptions.ConnectionError as ce:
                if "Failed to establish a new connection: [Errno -2] Name or service not known'" in str(ce):
                    logger.exception(str(ce))
                    logger.fatal("logger shut down, no access to cloud provider.")
                    Utils._reset_logger()
                    raise ce
                Utils._handle_exception(attempt_number, ce, logger, logger_msg, start, timeout_in_sec)
            except Exception as e:
                Utils._handle_exception(attempt_number, e, logger, logger_msg, start, timeout_in_sec)
            finally:
                time.sleep(interval)
            attempt_number += 1

    @staticmethod
    def _handle_exception(attempt_number, ce, logger, logger_msg, start, timeout_in_sec):
        elapsed = datetime.datetime.now() - start
        in_seconds = elapsed.total_seconds()
        if in_seconds > timeout_in_sec:
            if logger:
                logger.info("Failure while {} (timed out after {}s with {} attempts)"
                            .format(logger_msg, in_seconds, attempt_number))
            raise ce

    @staticmethod
    def _reset_logger():
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)
        logging.basicConfig()

    @staticmethod
    def convert_date_time_to_isoformat(date_time: datetime):
        return date_time.isoformat()

    @staticmethod
    def get_utc_now_in_isoformat() -> str:
        utc_now = datetime.datetime.now(datetime.timezone.utc)
        return Utils.convert_date_time_to_isoformat(utc_now)


class CallsLogger:
    logger = None

    @classmethod
    def set_logger(cls, logger):
        cls.logger = logger

    @classmethod
    def wrap(cls, func):
        def decorator(*args, **kwargs):
            bound_args = inspect.signature(func).bind(*args, **kwargs)
            bound_args.apply_defaults()

            args_list = ["{}: '{}'".format(k, v) for (k, v) in bound_args.arguments.items()]
            if len(bound_args.arguments) > 0 and bound_args.arguments.get('self'):
                args_list = args_list[1:]
            result_str = ''

            try:
                result = func(**bound_args.arguments)
                result_str = str(result)
                return result
            except Exception as e:
                result_str = str(e)
                raise
            finally:
                if cls.logger:
                    result_str = result_str if len(result_str) < 20 else result_str[:17]+'...'
                    cls.logger.info(func.__qualname__ + "(" + ", ".join(args_list) + ") -> "+result_str)

        return decorator



from logging import Logger
from typing import List
from sidecar.aws_session import AwsSession
from sidecar.const import Const
from sidecar.utils import Utils
from sidecar.status_maintainer import StatusMaintainer
from sidecar.sandbox_error import SandboxError


class AWSStatusMaintainer(StatusMaintainer):
    table_data = {}

    def __init__(self, awssessionservice: AwsSession, sandbox_id: str, logger: Logger, default_region: str,
                 table_name: str):
        super().__init__(logger)
        self.default_region = default_region
        self.dynamo_resource = awssessionservice.get_dynamo_resource(default_region=self.default_region)
        self.sandboxid = sandbox_id
        self._logger = logger
        self.table_name = table_name

        self.refresh_db()

    def _item_exist_in_db(self):
        table = self.dynamo_resource.Table(self.table_name)
        response = table.get_item(
            Key={
                Const.SANDBOX_ID_TAG: self.sandboxid
            }
        )
        return True if "Item" in response else False

    def _set_table_data(self, table, sandbox_id):
        response = table.get_item(
            Key={
                Const.SANDBOX_ID_TAG: sandbox_id
            }
        )

        if "Item" in response:
            self.table_data = response["Item"]
            self._logger.info("dynamo table response for sandbox {id} is {data}".format(id=sandbox_id, data=response))
            return "Item"

        return None

    def refresh_db(self):
        table = self.dynamo_resource.Table(self.table_name)
        try:
            Utils.wait_for(func=lambda: self._set_table_data(table, self.sandboxid) is not None,
                           interval_sec=1,
                           max_retries=5,
                           error='No sandbox data available')
        except Exception:
            self._logger.exception("Failed to get sandbox data from dynamodb after 5 times")
            return

    def update_qualiy_status(self, status: str):
        if not self._item_exist_in_db():
            return

        table = self.dynamo_resource.Table(self.table_name)
        response = table.update_item(
            Key={Const.SANDBOX_ID_TAG: self.sandboxid},
            UpdateExpression="set #f = :r",
            ExpressionAttributeValues={':r': status},
            ExpressionAttributeNames={"#f": Const.QUALIY_STATUS},
            ReturnValues="UPDATED_NEW"
        )

        if self.response_failed(response):
            raise Exception(f"Failed to update QualiY status '{status}' in sandbox '{self.sandboxid}'\n"
                            f"Response: {response}")

    def update_service_in_dynamo(self):
        if not self._item_exist_in_db():
            return

        table = self.dynamo_resource.Table(self.table_name)
        return table.update_item(
            Key={
                Const.SANDBOX_ID_TAG: self.sandboxid
            },
            UpdateExpression="set services = :r",
            ExpressionAttributeValues={
                ':r': self.table_data["services"]
            },
            ReturnValues="UPDATED_NEW"
        )

    def get_deployment_outputs(self, entity_name: str) -> {}:
        if entity_name in self.table_data["services"]:
            return self.table_data["services"][entity_name].get("outputs", {})

        for logical_id, logical_details in self.table_data["apps"].items():
            for instance_id, instance_json in logical_details['instances'].items():
                for app_name, app_details in instance_json['apps'].items():
                    if app_name == entity_name:
                        return app_details.get('outputs', {})
        raise Exception(f"could not find service/application with name '{entity_name}'")

    def get_service_deployment_outputs(self, service_name: str) -> {}:
        service = self.table_data["services"].get(service_name)
        if service:
            return service.get("outputs", {})
        raise Exception(f"could not find service with name '{service_name}'")

    def get_app_deployment_outputs(self, app_name: str) -> {}:
        app_details = self._get_first_application(app_name)
        if app_details:
            return app_details.get('outputs', {})
        raise Exception(f"could not find application with name '{app_name}'")

    def _get_first_application(self, app_name: str):
        for logical_id, logical_details in self.table_data["apps"].items():
            for instance_id, instance_json in logical_details['instances'].items():
                for name, app_details in instance_json['apps'].items():
                    if name == app_name:
                        return app_details
        return None

    def update_service_outputs(self, service_name: str, outputs: {}):
        if not self._item_exist_in_db():
            return

        self.table_data["services"][service_name]["outputs"] = outputs
        response = self.update_service_in_dynamo()
        if self.response_failed(response):
            self._logger.error(
                "error while updating service '{SERVICE}' outputs in sandbox '{SANDBOX}'. {RESPONSE}".format(
                    SERVICE=service_name,
                    SANDBOX=self.sandboxid,
                    RESPONSE=response))

    def update_app_instance_outputs(self, instance_logical_id, instance_id, app_name, outputs: {}):
        if not self._item_exist_in_db():
            return

        try:
            self.add_instance_to_data_table_if_not_exists(instance_id, instance_logical_id)
        except Exception as ex:  # log details for debugging, related to bug #1689
            self._logger.exception('fail to update instance outputs. instance_logical_id: {}, table_data: {}, exc={}'
                                   .format(instance_logical_id, self.table_data, str(ex)))
            raise ex

        app_details = self.table_data["apps"][instance_logical_id]["instances"][instance_id]["apps"] \
            .setdefault(app_name, {})

        app_details["outputs"] = outputs

        response = self.update_apps_in_dynamo()

        if self.response_failed(response):
            self._logger.error(
                "error while updating app instance outputs(sandbox_id: {sandbox_id}. app_name: "
                "{app}. instance_id: {instance_id})\n"
                "Response: {data}"
                    .format(sandbox_id=self.sandboxid, app=app_name, instance_id=instance_id, data=response))

    def update_service_status(self, name: str, status: str):
        if not self._item_exist_in_db():
            return

        self.table_data["services"][name]["status"] = status
        response = self.update_service_in_dynamo()
        if self.response_failed(response):
            self._logger.error(
                "error while updating service '{SERVICE}' status in sandbox '{SANDBOX}'. {RESPONSE}".format(
                    SERVICE=name,
                    SANDBOX=self.sandboxid,
                    RESPONSE=response))

    def update_app_instance_status(self, instance_logical_id, instance_id, app_name, status_tag, status):
        if not self._item_exist_in_db():
            return

        try:
            self.add_instance_to_data_table_if_not_exists(instance_id, instance_logical_id)
        except Exception as ex:  # log details for debugging, related to bug #1689
            self._logger.exception('fail to update instance status. instance_logical_id: {}, table_data: {}, exc={}'
                                   .format(instance_logical_id, self.table_data, str(ex)))
            raise ex

        app_details = self.table_data["apps"][instance_logical_id]["instances"][instance_id]["apps"]\
            .setdefault(app_name, {})

        app_details[status_tag] = status

        response = self.update_apps_in_dynamo()

        if self.response_failed(response):
            self._logger.error(
                "Error update_app_instance_status(sandbox_id: {sandbox_id}. app_name: {app}. instance_id: {instance_id})\n"
                "Response: {data}"
                    .format(sandbox_id=self.sandboxid, app=app_name, instance_id=instance_id, data=response))

    def update_apps_in_dynamo(self):
        if not self._item_exist_in_db():
            return

        table = self.dynamo_resource.Table(self.table_name)
        response = table.update_item(
            Key={
                Const.SANDBOX_ID_TAG: self.sandboxid
            },
            UpdateExpression="set " + "apps" + " = :r",
            ExpressionAttributeValues={
                ':r': self.table_data["apps"]
            },
            ReturnValues="UPDATED_NEW"
        )
        return response

    def add_instance_to_data_table_if_not_exists(self, instance_id, instance_logical_id):
        if instance_id not in self.table_data["apps"][instance_logical_id]["instances"]:
            self.table_data["apps"][instance_logical_id]["instances"][instance_id] = \
                {
                    "apps": {}
                }

    def update_lazy_load_artifacts_waiting(self, value: bool):
        if not self._item_exist_in_db():
            return

        table = self.dynamo_resource.Table(self.table_name)

        response = table.update_item(
            Key={
                Const.SANDBOX_ID_TAG: self.sandboxid
            },
            UpdateExpression="set #f = :r",
            ExpressionAttributeValues={
                ':r': value
            },
            ExpressionAttributeNames={
                "#f": Const.LAZY_LOAD_ARTIFACTS_WAITING
            },
            ReturnValues="UPDATED_NEW"
        )

        if self.response_failed(response):
            self._logger.error("Error update_lazy_load_artifacts_waiting(sandbox_id: {sandbox_id} value: {value})\n"
                               "Response: {data}"
                               .format(sandbox_id=self.sandboxid, value=value, data=response))

    def update_sandbox_errors(self, errors: List[SandboxError]):
        if not self._item_exist_in_db():
            return

        table = self.dynamo_resource.Table(self.table_name)

        errors_to_db = [{"message": error.message, "code": error.code, "time": error.time} for error in errors]

        response = table.update_item(
            Key={
                Const.SANDBOX_ID_TAG: self.sandboxid
            },
            UpdateExpression="set #f = :r",
            ExpressionAttributeValues={
                ':r': errors_to_db
            },
            ExpressionAttributeNames={
                "#f": Const.SANDBOX_ERRORS
            },
            ReturnValues="UPDATED_NEW"
        )

        if self.response_failed(response):
            self._logger.error("Error update_sandbox_errors(sandbox_id: {sandbox_id} errors: {errors})\n"
                               "Response: {data}"
                               .format(sandbox_id=self.sandboxid, errors=errors_to_db, data=response))


    def update_sandbox_end_status(self, sandbox_deployment_end_status: str):
        if not self._item_exist_in_db():
            return

        table = self.dynamo_resource.Table(self.table_name)

        response = table.update_item(
            Key={
                Const.SANDBOX_ID_TAG: self.sandboxid
            },
            UpdateExpression="set #f = :r, #newend = :r",
            ExpressionAttributeValues={
                ':r': sandbox_deployment_end_status
            },
            ExpressionAttributeNames={
                "#f": Const.SANDBOX_DEPLOYMENT_END_STATUS,
                "#newend": Const.SANDBOX_DEPLOYMENT_END_STATUS_v2
            },
            ReturnValues="UPDATED_NEW"
        )

        if self.response_failed(response):
            self._logger.error("Error update_sandbox_end_status(sandbox_id: {sandbox_id} status: {status})\n"
                               "Response: {data}"
                               .format(sandbox_id=self.sandboxid, status=sandbox_deployment_end_status, data=response))

    def update_sandbox_start_status(self, sandbox_start_time):
        if not self._item_exist_in_db():
            return

        table = self.dynamo_resource.Table(self.table_name)

        response = table.update_item(
            Key={
                Const.SANDBOX_ID_TAG: self.sandboxid
            },
            UpdateExpression="set #f = :r, #newstart = :r",
            ExpressionAttributeValues={
                ':r': str(sandbox_start_time)
            },
            ExpressionAttributeNames={
                "#f": Const.SANDBOX_START_TIME,
                "#newstart": Const.SANDBOX_START_TIME_v2
            },
            ReturnValues="UPDATED_NEW"
        )

        if self.response_failed(response):
            self._logger.error("Error update_sandbox_start_status(sandbox_id: {sandbox_id} status: {status})\n"
                               "Response: {data}"
                               .format(sandbox_id=self.sandboxid, status=sandbox_start_time, data=response))

    def get_all_app_names_for_instance(self, logical_id: str):
        return self.table_data["spec"]["expected_apps"][logical_id]["apps"]

    @staticmethod
    def response_failed(response: dict) -> bool:
        return not response.get("ResponseMetadata") and \
               not response.get("ResponseMetadata").get("HTTPStatusCode") == 200

    def get_ingress_routes(self) -> List['IngressRouteRequest']:
        items = []
        for ingress_route in self.table_data.get('ingress_routes', []):
            items.append(IngressRouteRequest(
                listener_port=int(ingress_route['listener_port']),
                path=ingress_route['path'],
                host=ingress_route['host'],
                app_name=ingress_route['app_name'],
                app_port=int(ingress_route['app_port']),
                color=ingress_route['color']))
        return items

    def get_sandbox_errors(self) -> List[SandboxError]:
        self.refresh_db()
        errors = []
        for error in self.table_data.get(Const.SANDBOX_ERRORS, []):
            errors.append(SandboxError(message=str(error["message"]),
                                       code=str(error["code"]),
                                       time=str(error["time"])))
        return errors

    def get_terminating_flag(self) -> bool:
        table = self.dynamo_resource.Table(self.table_name)

        response = table.get_item(
            Key={Const.SANDBOX_ID_TAG: self.sandboxid},
            ConsistentRead=True,
            ProjectionExpression='terminating'
        )

        if "Item" in response:
            doc = response["Item"]
            if 'terminating' in doc:
                return doc['terminating']

        return False


class IngressRouteRequest:
    def __init__(self, listener_port: int, path: str, host: str, app_port: int, app_name: str, color: str):
        self.listener_port = listener_port
        self.path = path
        self.host = host
        self.app_port = app_port
        self.app_name = app_name
        self.color = color



from logging import Logger
from typing import List

import datetime

from sidecar.app_instance_identifier import AppInstanceIdentifier
from sidecar.aws_instance_states import AwsInstanceStates
from sidecar.aws_session import AwsSession
from sidecar.aws_tag_helper import AwsTagHelper
from sidecar.const import Const
from sidecar.app_instance_service import IAppInstanceService, StaleAppInstanceException
from sidecar.aws_status_maintainer import AWSStatusMaintainer


class AwsAppInstanceService(IAppInstanceService):

    def __init__(self, sandbox_id: str, logger: Logger, aws_session: AwsSession,
                 aws_status_maintainer: AWSStatusMaintainer):
        super().__init__(logger)
        self.aws_session = aws_session
        self.sandbox_id = sandbox_id
        self.aws_status_maintainer = aws_status_maintainer

    def _update_app_status(self, instance_id: str, app_name: str, status: str):
        instance = self._get_existing_instance(instance_id, app_name)
        # should update status only for a "live" instance
        if not self._is_instance_part_of_sandbox(instance):
            raise StaleAppInstanceException("cannot update '{APP_NAME}' status to '{STATUS}' since the app instance is "
                                            "no longer a part of the sandbox. infra id: {INFRA_ID}"
                                            .format(APP_NAME=app_name, STATUS=status, INFRA_ID=instance_id))

        instance_logical_id = AwsTagHelper.wait_for_tag(instance, Const.INSTANCELOGICALID, self._logger)
        self.aws_status_maintainer.update_app_instance_status(instance_logical_id, instance_id, app_name,
                                                              Const.APP_STATUS_TAG, status)

    def get_deployment_outputs(self, app_name: str) -> {}:
        return self.aws_status_maintainer.get_app_deployment_outputs(app_name)

    def is_qualiy_off(self) -> bool:
        instance = self._get_qualiy_instance()
        if not instance:
            raise Exception('QualiY instance was not found')

        return instance.state["Name"] in [AwsInstanceStates.STOPPED, AwsInstanceStates.TERMINATED]

    def update_deployment_outputs(self, app_instance_identifier: AppInstanceIdentifier, outputs: {}):
        instance_id = app_instance_identifier.infra_id
        app_name = app_instance_identifier.name
        instance = self._get_existing_instance(instance_id, app_name)
        # should update status only for a "live" instance
        if not self._is_instance_part_of_sandbox(instance):
            raise StaleAppInstanceException("cannot update '{APP_NAME}' outputs since the app instance is "
                                            "no longer a part of the sandbox. infra id: {INFRA_ID}"
                                            .format(APP_NAME=app_name, INFRA_ID=instance_id))

        instance_logical_id = AwsTagHelper.wait_for_tag(instance, Const.INSTANCELOGICALID, self._logger)
        self.aws_status_maintainer.update_app_instance_outputs(instance_logical_id, instance_id, app_name, outputs)

    def _get_existing_instance(self, instance_id: str, app_name: str):
        instance = self._get_instance_by_id(instance_id)
        if instance is None:
            raise Exception("instance for app '{APP_NAME}' not found. instance_id={INSTANCE_ID}"
                            .format(APP_NAME=app_name, INSTANCE_ID=instance_id))
        return instance

    def get_public_address(self, app_instance_identifier: AppInstanceIdentifier) -> str:
        instance = self._get_existing_instance(instance_id=app_instance_identifier.infra_id,
                                               app_name=app_instance_identifier.name)
        alb_dns = next((tag['Value'] for tag in instance.tags if tag['Name'] == Const.EXTERNAL_ELB_DNS_NAME), None)
        return alb_dns

    def update_status_if_not_stale(self, app_instance_identifier: AppInstanceIdentifier, status: str):
        instance_id = app_instance_identifier.infra_id
        self._update_app_status(instance_id, app_instance_identifier.name, status)

    def check_which_exist(self, identifiers: List[AppInstanceIdentifier]) -> List[AppInstanceIdentifier]:
        instance_ids = list(set(identifier.infra_id for identifier in identifiers))
        instances = self._get_sandbox_app_instances_by_ids(instance_ids=instance_ids)

        existing_identifiers = [app_instance_identifier
                                for instance in instances
                                for app_instance_identifier in self._create_app_instance_identifiers(instance)
                                if app_instance_identifier in identifiers]
        return existing_identifiers

    def _get_instance_by_id(self, instance_id: str):
        ec2 = self.aws_session.get_ec2_resource()
        return ec2.Instance(instance_id)

    def _get_qualiy_instance(self):
        ec2 = self.aws_session.get_ec2_resource()

        filters = [{'Name': 'tag:' + Const.SANDBOX_ID_TAG,
                    'Values': [self.sandbox_id]},
                   {'Name': 'tag:' + Const.APP_NAME_TAG,
                    'Values': [Const.QUALY_SERVICE_NAME]}]
        instance = next(iter(list(ec2.instances.filter(Filters=filters))), None)
        return instance

    def _is_instance_part_of_sandbox(self, instance) -> bool:
        instance_as_list = [instance]
        self._filter_out_autoscaling_instances_not_in_asg(instance_as_list)
        return instance in instance_as_list

    def _get_sandbox_app_instances_by_ids(self, instance_ids: List[str]) -> list():
        instances = self._query_instances_by_ids(instance_ids)
        self._filter_live_app_instances(instances)
        return instances

    def _query_instances_by_ids(self, instance_ids: List[str]):
        instances = list(self.aws_session.get_ec2_resource().instances.filter(InstanceIds=instance_ids))
        self._logger.info("[{TIMESTAMP} utc] query for instances by ids ({queries_ids}) returned: {ids}"
                          .format(TIMESTAMP=datetime.datetime.now(datetime.timezone.utc).strftime('%X'),
                                  queries_ids=", ".join(instance_ids),
                                  ids=", ".join(inst.instance_id for inst in instances)))

        return instances

    def _filter_live_app_instances(self, instances: list()):
        self._filter_out_infra_instances(instances)
        self._filter_out_autoscaling_instances_not_in_asg(instances)

    def _filter_out_infra_instances(self, instances: list()):
        for instance in instances[:]:
            if self._is_infra_instance(instance):
                instances.remove(instance)

    def _filter_out_autoscaling_instances_not_in_asg(self, instances_to_filter: list()):
        asg_name_to_instances_map = self._get_asg_name_to_instances_map(instances_to_filter)
        asg_names = list(asg_name_to_instances_map.keys())
        if not asg_names:
            return
        asg_name_to_existing_instance_ids_map = self._get_existing_asg_instance_ids(asg_names)

        for asg_name, asg_instances_list in asg_name_to_instances_map.items():
            existing_instance_ids = asg_name_to_existing_instance_ids_map.get(asg_name, [])
            for instance in asg_instances_list:
                if instance.instance_id not in existing_instance_ids:
                    instances_to_filter.remove(instance)

    def _get_asg_name_to_instances_map(self, instances):
        asg_name_to_instances_map = dict()
        for instance in instances:
            asg_tag_value = AwsTagHelper.wait_for_tags(instance, self._logger).get(AwsTagHelper.AutoScalingGroupNameTag)
            if asg_tag_value:
                asg_instances_list = asg_name_to_instances_map.setdefault(asg_tag_value, [])
                asg_instances_list.append(instance)
        return asg_name_to_instances_map

    def _get_existing_asg_instance_ids(self, asg_names: List[str]) -> dict():
        if not asg_names:
            return {}
        autoscaling_groups = self._query_autoscaling_groups_by_names(asg_names)
        asg_name_to_existing_instance_ids_map = \
            {asg['AutoScalingGroupName']: [as_inst['InstanceId'] for as_inst in asg['Instances']]
             for asg in autoscaling_groups}
        return asg_name_to_existing_instance_ids_map

    def _query_autoscaling_groups_by_names(self, asg_names: List[str]) -> []:
        autoscaling_client = self.aws_session.get_autoscaling_client()
        response = autoscaling_client.describe_auto_scaling_groups(AutoScalingGroupNames=asg_names)
        autoscaling_groups = response['AutoScalingGroups']
        return autoscaling_groups

    def _create_app_instance_identifiers(self, instance) -> List[AppInstanceIdentifier]:
        instance_logical_id = AwsTagHelper.wait_for_tag(instance, Const.INSTANCELOGICALID, self._logger)
        all_app_names = self.aws_status_maintainer.get_all_app_names_for_instance(instance_logical_id)
        return [AppInstanceIdentifier(name=app_name, infra_id=instance.instance_id) for app_name in all_app_names]

    def _is_infra_instance(self, instance) -> bool:
        app_name_tag_value = AwsTagHelper.wait_for_tag(instance, Const.APP_NAME_TAG, self._logger)
        return app_name_tag_value in [Const.AWS_SIDECAR_APP_NAME, Const.QUALY_SERVICE_NAME]
